/*
 * $Id: midi.c,v 1.1.1.1 2002/11/28 09:02:09 eli7 Exp $
 */

/*  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static char *rcsid =
        "$Id: midi.c,v 1.1.1.1 2002/11/28 09:02:09 eli7 Exp $";

#include "midi.h"
#include "ll.h"
#include "pq.h"
#include "err.h"

#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <signal.h>

#define DEVICE	0

#define DEV "/dev/sequencer"
#define INSTRUMENT 60
#define DELTA		10 /* 10 ms */
#define MAX_CHANNEL 10
#ifndef VERBOSITY
#define VERBOSE(args...)
#else
#define VERBOSE(args...) BUGF(args)
#endif /* VERBOSITY */

typedef struct note_event {
	int velocity;
	ms time;
	int pitch;
	int on;
} note_event_t;

static ll_t pq;
static int seqfd = -1; /* must be the name of the seqfd fd! */
static struct timeval begin;
static int isopen = 0;

static int lock = 0;

static int channels[MAX_CHANNEL];

#define BUFFSIZE	1024
SEQ_DEFINEBUF(BUFFSIZE);

static void do_note(note_event_t *e);

static inline int 
get_id(note_event_t *e)
{
	return e->pitch * 128 + e->velocity;
}

static void 
init_channels() 
{
	memset(channels, 0, sizeof(channels));
}

static int 
get_channel(note_event_t *e) 
{

	int i, id;

	id = get_id(e);
	
	for (i = 0; i <	MAX_CHANNEL; i++) {
		if (channels[i] == 0) {
			break; // found empty
		}
	}
	if (i < MAX_CHANNEL) {
		channels[i] = id;
		return i + 1;
	} else {
		VERBOSE("can't play, no empty channels left");
		return -1;
	}
}

static int 
free_channel(note_event_t *e) 
{
	int i, id;

	id = get_id(e);
	
	for (i = 0; i < MAX_CHANNEL; i++) {
		if (channels[i] == id) {
			channels[i] = 0;
			return i + 1;
		}
	}
	BUGF("channel never allocated for id=%d!", id);
	return -1;
}

/*
int 
printer(void *data, void *args) {
	note_event_t *e = (note_event_t*) data;
	char *message = args;
	if (e != NULL) {
		printf("%s: BEEP(%d) (%d,%d) at %ld\n", message, e->on, e->pitch, e->velocity, 
				e->time);
   } else {
          printf("%s: NULL note!\n", message);
   }
   return 0;
}
*/

/*
 * locks
 */

static void 
spin_lock()
{
	while (lock) {
		// spin
	}
	lock = 1;
}

static void
spin_unlock()
{
	if (!lock) {
		BUGF("trying to unlock, but no lock.");
	}
	lock = 0;
}

/*
 * time functions
 */

static inline void
ms2tv(ms t, struct timeval *tv)
{
	tv->tv_sec = t / 1000;
	tv->tv_usec = (t % 1000) * 1000;
}

static inline ms
tv2ms(const struct timeval *tv)
{
	return tv->tv_sec * 1000 + (tv->tv_usec / 1000);
}

static void
init_midi_time()
{
	gettimeofday(&begin, NULL);
}

/*
 * alarm functions
 */

void alarm_handler(int);

static int 
init_alarm_handler()
{
	struct sigaction act;

	memset(&act, 0, sizeof(act));

	act.sa_handler = alarm_handler;
	if (signal(SIGALRM, alarm_handler) == SIG_ERR) {
		BUGF("sig error");
		return -1;
	}
	return 0;
}

static void
reset_alarm(note_event_t *e)
{
	struct itimerval itv;
	ms delta;

	memset(&itv, 0, sizeof(itv));
	
	if (e != NULL) {
		delta = e->time - midi_get_time();// FIXME inaccurate
		if (delta <= 0) {
			VERBOSE("attempted to set timer in %ldms.", delta);
			delta = 1; // just so we're not late
		}
		ms2tv(delta, &(itv.it_value));

		if ( setitimer(ITIMER_REAL, &itv, NULL) == -1) {
			BUGF(errnostr());
		}
	} else {
		VERBOSE("no events planned.");
	}
}

static void
alarm_die()
{
	if ( setitimer(ITIMER_REAL, NULL, NULL) == -1) {
		BUGF("couldnt unset alarm %s", errnostr());
	}
}

void
alarm_handler(int sig)
{
	void *e;
	spin_lock();
	if (pq_deq(&pq, &e) == -1) {
		BUGF("expect note, but empty queue.");
		goto alarm_handler_cleanup;
	} 
	do_note(e);
	if (pq_topq(&pq, &e) == -1) {
		// last event, empty queue
		goto alarm_handler_cleanup;
	}
	reset_alarm((note_event_t*) e);

alarm_handler_cleanup:
	spin_unlock();
}

int
notecmp(void *p1, void *p2)
{
	note_event_t *n1, *n2;
	
	n1 = (note_event_t*) p1;
	n2 = (note_event_t*) p2;

	return (n1->time - n2->time);
}

static void                           
add_note_event(note_event_t *e)
{
	void *topq;
	spin_lock();
	pq_enq(&pq, e, notecmp);
	pq_topq(&pq, &topq);
	
	if (e == topq) { // yes, the actual pointers
		reset_alarm(e);
	}
	spin_unlock();
}

static void
free_events()
{
	note_event_t *e;
	while (pq_deq(&pq, (void**) &e) != -1) {
		free(e); 
	}
}

/*
 * midi stuff
 */

/*
 * NOTE! Don't ever define SEQ_DUMPBUF() in two source files or when OSSlib
 *       is used. It may have unpredictable results.
 */
#ifndef OSSLIB
void
seqbuf_dump()
{
	if (_seqbufptr) {
		if (write(seqfd, _seqbuf, _seqbufptr) == -1) {
			BUGF("DEV: %s",errnostr());
		}
	}
  	_seqbufptr = 0;
}
#endif /* OSSLIB */

static int 
init_devices(int instrument)
{
	int channel;
	
	SEQ_LOAD_GMINSTR(DEVICE, instrument); 
	SEQ_START_TIMER();
	
	for (channel = 1; channel <= MAX_CHANNEL; channel++) {
		SEQ_PGM_CHANGE(DEVICE, channel, instrument);
		SEQ_START_NOTE(DEVICE, channel, 70 + channel*2, 40); 
	}
	SEQ_DUMPBUF();                      /* flush buf */

	sleep(1);
	
	for (channel = 1; channel <= MAX_CHANNEL; channel++) {
		SEQ_STOP_NOTE(DEVICE, channel, 70 + channel*2, 40);
	}
	SEQ_DUMPBUF();                      /* flush buf */
	
	return 0;
}

static int 
init_midi() 
{
	int ndevices;
	

	if ( (seqfd = open(DEV, O_WRONLY, 0)) == -1) {
		VERBOSE("%s: %s", DEV, errnostr());
		return -1;
	}
	else if (ioctl(seqfd, SNDCTL_SEQ_NRSYNTHS, &ndevices)==-1)
	{
		VERBOSE("SNDCTL_SEQ_NRSYNTHS: %s: %s", DEV, errnostr());
		goto init_midi_cleanup;
	}

	else if (DEVICE >= ndevices) {
		VERBOSE("playback dev %d (< %d) doesn't exist\n", DEVICE, ndevices);
		goto init_midi_cleanup;
	} else if (init_devices(INSTRUMENT) == -1) {
			goto init_midi_cleanup;
	} else {
		return 0;
	}

init_midi_cleanup:
	close(seqfd);
	return -1;
}

static void 
do_note(note_event_t *e)
{
	int late, channel;

	if (e == NULL) {
	 	BUGF("null note event!");
		return;
	}	 
	
	late = midi_get_time() - e->time;
	
	if (abs(late) > DELTA) {
		VERBOSE("Alarm late in %dlms.", late);
	}
	
	if (e->on) {
		
		if ( (channel = get_channel(e)) == -1) {
			return;
		}
		SEQ_START_NOTE(DEVICE, channel, e->pitch, e->velocity);
		SEQ_DUMPBUF();
		VERBOSE("BEEP on %d (%d,%d)", channel, e->pitch, e->velocity);
	} else {
		if ( (channel = free_channel(e)) == -1) {
			return;
		}
		
		SEQ_STOP_NOTE(DEVICE, channel, e->pitch, e->velocity);
		SEQ_DUMPBUF();
		VERBOSE("shhh on %d (%d,%d)", channel, e->pitch, e->velocity);
	}
	free(e);
}


static int 
midi_die() {

	SEQ_DELTA_TIME(1000); 		/* ticks for music decay */
	SEQ_DUMPBUF();                      /* flush buf */
	if (close(seqfd) == -1) {
		VERBOSE("%s: %s", DEV, errnostr());
		return 0;
	}
	return 1;
}


/*
 * scheme interface!
 */

ms
midi_get_time()
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	tv.tv_sec -= begin.tv_sec;
	tv.tv_usec -= begin.tv_usec;
	return tv2ms(&tv);
}

void
midi_close() 
{
	if (isopen) {
		alarm_die();
		free_events();
		ll_die(&pq);
		midi_die();
		isopen = 0;
	}
}


int
midi_open()
{
	if (isopen) {
		midi_close();
	}
	init_channels();
	init_midi_time();
	if (init_midi() == -1) {
		goto midi_open_cleanup;
	} else if (ll_init(&pq) == -1) {
		goto midi_open_midi_cleanup;
	} else if (init_alarm_handler() == -1) {
		goto midi_open_ll_cleanup;
        } else {
		isopen = 1;
		sleep(1);
		VERBOSE("MIDI open");
		return 0;
	}

midi_open_ll_cleanup:	ll_die(&pq);
midi_open_midi_cleanup:	midi_die();
midi_open_cleanup:	VERBOSE("midi_open failed");
	return -1;
	
	
}

int
play_note(int pitch, ms duration, int velocity, ms midi_time)
{
	note_event_t *onset; 
	note_event_t *offset; 

	if (!isopen) {
		VERBOSE("midi lib not open!");
		return -1;
	}
	if (duration <= 0 || velocity <= 0 || midi_time <= 0) {
		return -1;
	}
	onset = malloc(sizeof(note_event_t));
	offset = malloc(sizeof(note_event_t));  
	if (offset == NULL || onset == NULL){
		return -1;
	}
	offset->time = duration + (onset->time = midi_time);
	offset->velocity = onset->velocity = velocity;
	offset->pitch = onset->pitch = pitch;
	offset->on = 0;
	onset->on = 1;
	add_note_event(onset);
	add_note_event(offset);
	return 0;
}


